from zope.interface import (
    Interface,
    Attribute,
    )

class IIndexInjection(Interface):
    """Interface for injecting documents into an index."""

    def index_doc(docid, value):
        """Add a document to the index.

        docid: int, identifying the document

        value: the (undiscriminated) value to be indexed

        return: None

        This can also be used to reindex documents.
        """

    def unindex_doc(docid):
        """Remove a document from the index.

        docid: int, identifying the document

        return: None

        This call is a no-op if the docid isn't in the index, however,
        after this call, the index should have no references to the docid.
        """

    def reindex_doc(docid, value):
        """Reindex a document using the (undiscriminated) value"""

    def reset():
        """Unindex all documents indexed by the index
        """

class IIndexEnumeration(Interface):
    def not_indexed():
        """ Return the set of document ids for which this index's
        discriminator returned ``default`` during ``index_doc`` or
        ``reindex_doc``, indicating that the index configuration was
        uninterested in indexing the value."""

    def not_indexed_count():
        """Return the number of document ids currently not indexed.
        Logically equivalent to len(self.not_indexed())."""

    def indexed():
        """ Return the set of document ids for which this index's
        discriminator returned a non-default value during ``index_doc`` or
        ``reindex_doc``, indicating that the index configuration as
        interested in indexing the value."""

    def indexed_count():
        """Return the number of document ids currently indexed.  Logically
        equivalent to len(self.indexed())."""

    def docids():
        """ Return the set of document ids which have been reported to this
        index via its ``index_doc`` or ``reindex_doc`` method (including
        document ids which had values which were not indexed).  This is the
        logical union of sets returned by indexed() and not_indexed()."""

    def docids_count():
        """Return the number of document ids currently in the set of both
        indexed and not indexed.  Logically equivalent to
        len(self.docids())."""

class IIndexDisplay(Interface):
    def qname():
        """ Return a string suitable for representing the index's name """

    def document_repr(docid, default=None):
        """ Return an index-specific string representation for the docid value,
        or the default if no such docid exists in the index."""

class IIndex(IIndexInjection, IIndexEnumeration, IIndexDisplay):
    pass

class IIndexStatistics(Interface):
    """An index that provides statistical information about itself."""

    def word_count():
        """Return the number of words currently indexed."""

class IIndexSort(Interface):

    def sort(docids, reverse=False, limit=None, sort_type=None,
             raise_unsortable=True):
        """Sort document ids sequence using indexed values
        
        Return a sorted iterable of document ids. Limited by value of the
        'limit' argument and optionally reversed, using the 'reverse'
        argument.

        If ``sort_type`` is not ``None``, it should be the value
        :attr:`hypatia.interfaces.STABLE` to specify that the sort should be
        stable (see http://www.algorithmist.com/index.php/Stable_Sort) or
        :attr:`hypatia.interfaces.OPTIMAL` to specify that the sort should be
        done with the opmimal algorithm (which is not guaranteed to be stable).
        By default, the sort will be done using the optimal algorithm if
        ``sort_type`` is not :attr:`~hypatia.interfaces.STABLE``.

        If ``raise_unsortable`` is ``True`` (the default), if the index cannot
        resolve any of the docids in the set of docids in this result set, a
        :exc:`hypatia.exc.Unsortable` exception will be raised during iteration
        over the sorted docids.  If ``raise_unsortable`` is ``False``, if some
        of docids are not indexed they are skipped from resulting iterable.
        """

class ICatalog(IIndexInjection):
    """ Dictionary-like object which maps index names to index instances.
    Also supports the IIndexInjection interface."""

class ICatalogQuery(Interface):
    def __call__(queryobject, sort_index=None, limit=None, sort_type=None,
                 reverse=False, names=None):
        """Search the catalog using the query and options provided.  """

class IResultSet(Interface):
    """ Iterable sequence of documents or document identifiers."""

    ids = Attribute('An iterable sequence of document identifiers')

    resolver = Attribute(
        'A callable which accepts a document id and which returns a document.  '
        'May be ``None``, in which case, resolution performed by result set '
        'methods is not performed, and document identifiers are returned '
        'unresolved.'
        )

    def __len__():
        """ Return the length of the result set"""

    def sort(index, reverse=False, limit=None, sort_type=None,
             raise_unsortable=True):
        """Return another IResultSet sorted using the ``index`` (an IIndexSort)
        passed to it after performing the sort using the index and the
        ``limit``, ``reverse``, and ``sort_type`` parameters.

        If ``sort_type`` is not ``None``, it should be the value
        :attr:`hypatia.interfaces.STABLE` to specify that the sort should be
        stable or :attr:`hypatia.interfaces.OPTIMAL` to specify that the sort
        algorithm chosen should be optimal (but not necessarily stable).  It's
        usually unnecessary to pass this value, even if you're resorting an
        already-sorted set of docids, because the implementation of IResultSet
        will internally ensure that subsequent sorts of the returned result set
        of an initial sort will be stable; if you don't want this behavior,
        explicitly pass :attr:`hypatia.interfaces.OPTIMAL` on the second and
        subsequent sorts of a set of docids.

        If ``raise_unsortable`` is ``True`` (the default), if the index cannot
        resolve any of the docids in the set of docids in this result set, a
        :exc:`hypatia.exc.Unsortable` exception will be raised during iteration
        over the sorted docids.
        """

    def first(resolve=True):
        """ Return the first element in the sequence.  If ``resolve`` is True,
        and the result set has a valid resolver, return the resolved
        document, otherwise return the document id of the first document. """

    def one(resolve=True):
        """ Return the element in the resultset, asserting that there is only
        one result.  If the resultset has more than one element, raise an
        :exc:`hypatia.exc.MultipleResults` exception.  If the resultset has no
        elements, raise an :exc:`hypatia.exc.NoResults` exception. `If
        ``resolve`` is True, and the result set has a valid resolver, return
        the resolved document, otherwise return the document id of the
        document."""

    def all(resolve=True):
        """ Return a sequence representing all elements in the resultset.  `If
        ``resolve`` is True, and the result set has a valid resolver, return an
        iterable of the resolved documents, otherwise return an iterable
        containing the document id of each document."""

    def __iter__():
        """ Return an iterator over the results of ``self.all()``"""
        
FWSCAN = 'fwscan'
NBEST = 'nbest'
TIMSORT = 'timsort'
STABLE = 'stable'
OPTIMAL = 'optimal'



import unittest
from .. import query

class _CatalogMaker(object):
    def _makeCatalog(self):
        from ..catalog import Catalog
        from ..field import FieldIndex
        from ..keyword import KeywordIndex
        from ..text import TextIndex

        catalog = Catalog()

        self.name = FieldIndex('name')
        self.title = FieldIndex('title')
        self.text = TextIndex('text')
        self.allowed = KeywordIndex('allowed')

        catalog['name'] = self.name
        catalog['title'] = self.title
        catalog['text'] = self.text
        catalog['allowed'] = self.allowed

        catalog.index_doc(1, Content('name1', 'title1', 'body one', ['a']))
        catalog.index_doc(2, Content('name2', 'title2', 'body two', ['b']))
        catalog.index_doc(3, Content('name3', 'title3', 'body three', ['c']))
        catalog.index_doc(4, Content('name4', None, 'body four',['a', 'b']))
        catalog.index_doc(5, Content('name5', 'title5', 'body five',
                                     ['a', 'b', 'c']))
        catalog.index_doc(6, Content('name6', 'title6', 'body six',['d']))

        return catalog

class TestCatalogQueryBase(_CatalogMaker):

    def test_it(self):
        from ..catalog import CatalogQuery

        catalog = self._makeCatalog()

        q = CatalogQuery(catalog)

        numdocs, result = q.query(
            self.get_query(),
            sort_index='name',
            limit=5,
            names=dict(body='body')
            )

        self.assertEqual(numdocs, 2)
        self.assertEqual(list(result), [4, 5])


class TestCatalogQueryWithCQE(unittest.TestCase, TestCatalogQueryBase):
    def get_query(self):
        return (
            "(allowed == 'a' and allowed == 'b' and "
            "(name == 'name1' or name == 'name2' or name == 'name3' or "
            "name == 'name4' or name == 'name5') and not(title == 'title3')) "
            "and body in text"
            )

class TestCatalogQueryWithPythonQueryObjects(
    unittest.TestCase,
    TestCatalogQueryBase
    ):
    def get_query(self):
        return (
            query.All(self.allowed, ['a', 'b']) &
            query.Any(self.name,
                      ['name1', 'name2', 'name3', 'name4', 'name5']) &
            query.Not(query.Eq(self.title, 'title3')) &
            query.Contains(self.text, 'body')
            )

class TestCatalogQueryWithPythonQueryObjectsUnionAndIntersection(
    unittest.TestCase,
    _CatalogMaker,
    ):
    def _get_result(self, q):
        from ..catalog import CatalogQuery

        cq = CatalogQuery(self.catalog)

        numdocs, result = cq.query(
            q,
            sort_index='name',
            limit=5,
            names=dict(body='body')
            )
        return numdocs, result
    
    def test_union_both_left_and_right_nonempty(self):
        self.catalog = self._makeCatalog()
        q = (
            query.All(self.allowed, ['a', 'b']) &
            query.Any(self.name,
                      ['name1', 'name2', 'name3', 'name4', 'name5']) &
            query.Not(query.Eq(self.title, 'title3')) |
            query.Contains(self.text, 'body')
            )
        numdocs, result = self._get_result(q)
        self.assertEqual(numdocs, 5)
        self.assertEqual(list(result), [1, 2, 3, 4, 5])

    def test_union_left_nonempty_right_empty(self):
        self.catalog = self._makeCatalog()
        q = (
            query.All(self.allowed, ['a', 'b']) &
            query.Any(self.name,
                      ['name1', 'name2', 'name3', 'name4', 'name5']) &
            query.Not(query.Eq(self.title, 'title3')) |
            query.Contains(self.text, 'nope')
            )
        numdocs, result = self._get_result(q)
        self.assertEqual(numdocs, 2)
        self.assertEqual(list(result), [4, 5])

    def test_union_left_empty_right_nonempty(self):
        self.catalog = self._makeCatalog()
        q = (
            query.Contains(self.text, 'nope') |
            query.All(self.allowed, ['a', 'b']) &
            query.Any(self.name,
                      ['name1', 'name2', 'name3', 'name4', 'name5']) &
            query.Not(query.Eq(self.title, 'title3'))
            )
        numdocs, result = self._get_result(q)
        self.assertEqual(numdocs, 2)
        self.assertEqual(list(result), [4, 5])

    def test_intersection_left_nonempty_right_empty(self):
        self.catalog = self._makeCatalog()
        q = (
            query.Contains(self.text, 'body') &
            query.All(self.allowed, ['c', 'd'])
            )
        numdocs, result = self._get_result(q)
        self.assertEqual(numdocs, 0)
        self.assertEqual(list(result), [])
        
class TestCatalogQueryWithIndexHelpers(
    unittest.TestCase,
    TestCatalogQueryBase
    ):
    def get_query(self):
        return (
            self.allowed.all(['a', 'b']) &
            self.name.any(['name1', 'name2', 'name3', 'name4', 'name5']) &
            self.title.noteq('title3') &
            self.text.contains('body')
            )

class TestQueryExecution(unittest.TestCase, _CatalogMaker):
    def test_it(self):
        self._makeCatalog()
        query = (
            self.allowed.all(['a', 'b']) &
            self.name.any(['name1', 'name2', 'name3', 'name4', 'name5']) &
            self.title.noteq('title3') &
            self.text.contains('body')
            )
        resultset = query.execute().sort(self.name, limit=5)
        self.assertEqual(len(resultset), 2)
        self.assertEqual(list(resultset.ids), [4, 5])

class TestFieldIndexResultSetSortStabilityGuarantee(unittest.TestCase):
    def _makeCatalog(self):
        from ..catalog import Catalog
        from ..field import FieldIndex

        catalog = Catalog()

        name = FieldIndex('name')
        title = FieldIndex('title')

        catalog['name'] = name
        catalog['title'] = title

        catalog.index_doc(1, Content('name1', 'title1', None, None))
        catalog.index_doc(2, Content('name2', 'title1', None, None))
        catalog.index_doc(3, Content('name3', 'title2', None, None))
        catalog.index_doc(4, Content('name4', 'title2', None, None))
        catalog.index_doc(5, Content('name5', 'title2', None, None))
        catalog.index_doc(6, Content('name6', 'title3', None, None))

        return catalog

    def test_it(self):
        catalog = self._makeCatalog()
        name_index = catalog['name']
        title_index = catalog['title']
        query = name_index.any(
            ['name6', 'name5', 'name4', 'name3', 'name2', 'name1']
            )
        # first we sort by name, which will give us 1,2,3,4,5,6
        resultset = query.execute().sort(name_index)
        # then we sort the resulting result set by title; the ordering should
        # not change (even though some of the titles are identical, and might
        # be considered interchangeable) because the 2nd and subsequent sorts
        # of a result set are implicitly marked as stable.
        resultset = resultset.sort(title_index)
        self.assertEqual(len(resultset), 6)
        self.assertEqual(list(resultset.ids), [1, 2, 3, 4, 5, 6])

class Content(object):
    def __init__(self, name, title, text, allowed):
        self.name = name
        if title:
            self.title = title
        self.text = text
        self.allowed = allowed





import unittest

from . import RichComparisonMixin

_marker = object()

class TestResultSet(unittest.TestCase):
    def _getTargetClass(self):
        from . import ResultSet
        return ResultSet
        
    def _makeOne(self, ids, numids, resolver):
        cls = self._getTargetClass()
        return cls(ids, numids, resolver)

    def test_class_implements_IResultSet(self):
        from zope.interface.verify import verifyClass
        from hypatia.interfaces import IResultSet
        verifyClass(IResultSet, self._getTargetClass())

    def test_instance_implements_IResultSet(self):
        from zope.interface.verify import verifyObject
        from hypatia.interfaces import IResultSet
        inst = self._makeOne([1], 1, None)
        verifyObject(IResultSet, inst)
        
    def test___len__(self):
        inst = self._makeOne([1], 1, None)
        self.assertEqual(len(inst), 1)

    def test_sort(self):
        inst = self._makeOne([2, 1], 2, None)
        index = DummyIndex()
        result = inst.sort(index)
        self.assertEqual(result.ids, [1,2])
        self.assertEqual(result.numids, 2)
        self.assertEqual(result.resolver, None)
        self.assertEqual(index.reverse, False)
        self.assertEqual(index.limit, None)

    def test_sort_withargs(self):
        inst = self._makeOne([2, 1], 2, None)
        index = DummyIndex()
        result = inst.sort(index, limit=1, reverse=True)
        self.assertEqual(result.ids, [1,2])
        self.assertEqual(result.numids, 1)
        self.assertEqual(result.resolver, None)
        self.assertEqual(index.reverse, True)
        self.assertEqual(index.limit, 1)

    def test_sort_generator(self):
        def mygen():
            yield 2
            yield 1
            yield 3
        inst = self._makeOne(mygen(), 3, None)
        index = DummyIndex()
        result = inst.sort(index)
        self.assertEqual(result.ids, [1, 2, 3])
        self.assertEqual(result.numids, 3)

    def test_second_sort_stable(self):
        from hypatia.interfaces import STABLE, OPTIMAL
        inst = self._makeOne([2, 1], 2, None)
        index = DummyIndex()
        result = inst.sort(index, limit=1, reverse=True)
        self.assertEqual(result.ids, [1,2])
        self.assertEqual(result.numids, 1)
        self.assertEqual(result.resolver, None)
        self.assertEqual(index.reverse, True)
        self.assertEqual(index.limit, 1)
        self.assertEqual(index.sort_type, None)
        result2 = result.sort(index, limit=2)
        self.assertEqual(result2.ids, [1,2])
        self.assertEqual(result2.numids, 1)
        self.assertEqual(result2.resolver, None)
        self.assertEqual(index.reverse, False)
        self.assertEqual(index.limit, 2)
        self.assertEqual(index.sort_type, STABLE)
        result3 = result2.sort(index, limit=3, sort_type=OPTIMAL)
        self.assertEqual(result3.ids, [1,2])
        self.assertEqual(result3.numids, 1)
        self.assertEqual(result3.resolver, None)
        self.assertEqual(index.reverse, False)
        self.assertEqual(index.limit, 3)
        self.assertEqual(index.sort_type, OPTIMAL)

    def test_first_no_docids(self):
        ids = []
        inst = self._makeOne(ids, 0, None)
        self.assertEqual(inst.first(), None)
        self.assertEqual(inst.ids, ids)

    def test_first_idempotent_with_list_noresolve(self):
        ids = [2, 1]
        inst = self._makeOne(ids, 2, None)
        self.assertEqual(inst.first(), 2)
        self.assertEqual(inst.first(), 2)
        self.assertEqual(inst.ids, ids)

    def test_first_idempotent_with_tuple_noresolve(self):
        ids = (2, 1)
        inst = self._makeOne(ids, 2, None)
        self.assertEqual(inst.first(), 2)
        self.assertEqual(inst.first(), 2)
        self.assertEqual(inst.ids, ids)
        
    def test_first_idempotent_with_generator_noresolve(self):
        def gen():
            for x in (2, 1):
                yield x
        ids = gen()
        inst = self._makeOne(ids, 2, None)
        self.assertEqual(inst.first(), 2)
        self.assertEqual(inst.first(), 2)
        self.assertNotEqual(inst.ids, ids)

    def test_first_idempotent_with_list(self):
        def resolver(val):
            return val
        ids = [2, 1]
        inst = self._makeOne(ids, 2, resolver)
        self.assertEqual(inst.first(), 2)
        self.assertEqual(inst.first(), 2)
        self.assertEqual(inst.ids, ids)

    def test_first_idempotent_with_tuple(self):
        def resolver(val):
            return val
        ids = (2, 1)
        inst = self._makeOne(ids, 2, resolver)
        self.assertEqual(inst.first(), 2)
        self.assertEqual(inst.first(), 2)
        self.assertEqual(inst.ids, ids)
        
    def test_first_idempotent_with_generator(self):
        def resolver(val):
            return val
        def gen():
            for x in (2, 1):
                yield x
        ids = gen()
        inst = self._makeOne(ids, 2, resolver)
        self.assertEqual(inst.first(), 2)
        self.assertEqual(inst.first(), 2)
        self.assertNotEqual(inst.ids, ids)
        
    def test_first_doesnt_affect_iterability(self):
        ids = [2, 1]
        inst = self._makeOne(ids, 2, None)
        self.assertEqual(inst.first(), 2)
        self.assertEqual(list(inst), [2,1])
        
    def test_first_resolve_true_no_resolver(self):
        ids = [2, 1]
        inst = self._makeOne(ids, 2, None)
        self.assertEqual(inst.first(), 2)

    def test_first_resolve_true_with_resolver(self):
        def resolver(val):
            return 'a'
        inst = self._makeOne([2, 1], 2, resolver)
        self.assertEqual(inst.first(), 'a')

    def test_first_resolve_false(self):
        def resolver(val): return 'a'
        inst = self._makeOne([2, 1], 2, resolver)
        self.assertEqual(inst.first(resolve=False), 2)

    def test_one_no_docids(self):
        from ..exc import NoResults
        inst = self._makeOne([], 0, None)
        self.assertRaises(NoResults, inst.one)

    def test_one_more_than_one_docid(self):
        from ..exc import MultipleResults
        inst = self._makeOne([1, 2], 2, None)
        self.assertRaises(MultipleResults, inst.one)

    def test_one_resolve_true_no_resolver(self):
        inst = self._makeOne([1], 1, None)
        self.assertEqual(inst.one(), 1)

    def test_one_resolve_true_with_resolver(self):
        def resolver(val):
            return 'a'
        inst = self._makeOne([1], 1, resolver)
        self.assertEqual(inst.one(), 'a')

    def test_one_resolve_false(self):
        def resolver(val): return 'a'
        inst = self._makeOne([1], 1, resolver)
        self.assertEqual(inst.one(resolve=False), 1)

    def test_all_resolve_false(self):
        inst = self._makeOne([2, 1], 2, None)
        self.assertEqual(inst.all(resolve=False), [2, 1])

    def test_all_resolve_true_no_resolver(self):
        inst = self._makeOne([2, 1], 2, None)
        self.assertEqual(inst.all(), [2, 1])

    def test_all_resolve_true_with_resolver(self):
        def resolver(val):
            return 'a'
        inst = self._makeOne([2, 1], 2, resolver)
        self.assertEqual(list(inst.all()), ['a', 'a'])

    def test___iter__(self):
        def resolver(val):
            return 'a'
        inst = self._makeOne([2, 1], 2, resolver)
        self.assertEqual(list(iter(inst)), ['a', 'a'])

    def test_intersect_docids(self):
        inst = self._makeOne([3, 2, 1], 3, None)
        result = inst.intersect([1])
        self.assertEqual(result.__class__, inst.__class__)
        self.assertEqual(result.ids, [1])

    def test_intersect_rs(self):
        inst = self._makeOne([3, 2, 1], 3, None)
        inst2 = self._makeOne([3, 2], 2, None)
        result = inst.intersect(inst2)
        self.assertEqual(result.__class__, inst.__class__)
        self.assertEqual(result.ids, [3, 2])
        
class TestBaseIndexMixin(unittest.TestCase):
    def _getTargetClass(self):
        from . import BaseIndexMixin
        return BaseIndexMixin

    def _makeIndex(self, discriminator):
        import BTrees
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        index = Test()
        index.discriminator = discriminator
        index._docids = BTrees.family64.IF.Set()
        index._not_indexed = BTrees.family64.IF.TreeSet()
        index.family = BTrees.family64
        return index

    def test_index_doc_callback_returns_nondefault(self):
        def callback(ob, default):
            return ob
        index = self._makeIndex(callback)
        self.assertEqual(index.index_doc(1, 'abc'), 'abc')
        self.assertEqual(index.value, 'abc')
        self.assertEqual(set(index.docids()), set([1]))

    def test_index_doc_string_discrim(self):
        index = self._makeIndex('abc')
        class Dummy:
            abc = 'abc'
        dummy = Dummy()
        self.assertEqual(index.index_doc(1, dummy), 'abc')
        self.assertEqual(index.value, 'abc')
        self.assertEqual(set(index.docids()), set([1]))

    def test_index_doc_missing_value_unindexes(self):
        index = self._makeIndex('abc')
        class Dummy:
            pass
        dummy = Dummy()
        dummy.abc = 'abc'
        self.assertEqual(index.index_doc(1, dummy), 'abc')
        del dummy.abc
        del index.value
        self.assertEqual(index.index_doc(1, dummy), None)
        self.assertEqual(set(index.docids()), set([1]))
        self.assertEqual(index.value, None)
        self.assertEqual(index.unindexed, 1)

    def test_index_doc_missing_value_then_with_value(self):
        index = self._makeIndex('abc')
        class Dummy:
            pass
        dummy = Dummy()
        self.assertEqual(index.index_doc(20, dummy), None)
        self.assertTrue(20 in index.docids())
        dummy.abc = 'foo'
        self.assertEqual(index.index_doc(20, dummy), 'foo')
        self.assertTrue(20 in index.docids())

    def test_index_doc_missing_value_then_unindex(self):
        index = self._makeIndex('abc')
        class Dummy:
            pass
        dummy = Dummy()
        self.assertEqual(index.index_doc(20, dummy), None)
        self.assertTrue(20 in index.docids())
        index.unindex_doc(20)
        self.assertFalse(20 in index.docids())

    def test_docids_with_indexed_and_not_indexed(self):
        index = self._makeIndex('abc')
        class Dummy:
            pass
        dummy = Dummy()
        self.assertEqual(index.index_doc(20, dummy), None)
        dummy.abc = 'foo'
        self.assertEqual(index.index_doc(21, dummy), 'foo')
        docids = index.docids()
        self.assertEqual(set(docids), set([20, 21]))

    def test_indexed_count(self):
        inst = self._makeIndex('abc')
        def indexed(): return [1,2,3]
        inst.indexed = indexed
        self.assertEqual(inst.indexed_count(), 3)

    def test_not_indexed_count(self):
        inst = self._makeIndex('abc')
        def not_indexed(): return [1,2,3]
        inst.not_indexed = not_indexed
        self.assertEqual(inst.not_indexed_count(), 3)

    def test_docids_count(self):
        inst = self._makeIndex('abc')
        def docids(): return [1,2,3]
        inst.docids = docids
        self.assertEqual(inst.docids_count(), 3)

    def test_index_doc_persistent_value_raises(self):
        from persistent import Persistent
        index = self._makeIndex('abc')
        index._docids = set()
        class Dummy:
            pass
        dummy = Dummy()
        dummy.abc = Persistent()
        self.assertRaises(ValueError, index.index_doc, 1, dummy)

    def test_index_doc_broken_object_raises(self):
        from ZODB.broken import Broken
        index = self._makeIndex('abc')
        index._docids = set()
        class Dummy:
            pass
        dummy = Dummy()
        dummy.abc = Broken()
        self.assertRaises(ValueError, index.index_doc, 1, dummy)

    def test_reindex_doc(self):
        index = self._makeIndex('abc')
        index._docids = set()
        class Dummy:
            abc = 'abc'
        dummy = Dummy()
        index.index_doc(1, dummy)
        index.reindex_doc(1, dummy)
        self.assertEqual(index.unindexed, 1)
        self.assertEqual(index.value, 'abc')
        self.assertEqual(set(index.docids()), set([1]))

    def test_qname_with___name__(self):
        index = self._makeIndex('abc')
        index.__name__ = 'fred'
        self.assertEqual(index.qname(), 'fred')

    def test_qname_without___name__(self):
        index = self._makeIndex('abc')
        self.assertEqual(index.qname(), str(index))

    def test_flush(self):
        index = self._makeIndex('abc')
        self.assertEqual(index.flush(), None)

class RichComparisonMixinTest(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.comp = RichComparer(6)

    def testDefaultComparison(self):
        self.assertRaises(NotImplementedError,
                          lambda: RichComparisonMixin() == 3)
        self.assertRaises(NotImplementedError,
                          lambda: RichComparisonMixin() != 3)
        self.assertRaises(NotImplementedError,
                          lambda: RichComparisonMixin() < 3)
        self.assertRaises(NotImplementedError,
                          lambda: RichComparisonMixin() > 3)
        self.assertRaises(NotImplementedError,
                          lambda: RichComparisonMixin() <= 3)
        self.assertRaises(NotImplementedError,
                          lambda: RichComparisonMixin() >= 3)


    def testEquality(self):
        self.assertTrue(self.comp == 6)
        self.assertTrue(self.comp == RichComparer(6))

        self.assertFalse(self.comp == 7)
        self.assertFalse(self.comp == RichComparer(7))

    def testInEquality(self):
        self.assertFalse(self.comp != 6)
        self.assertFalse(self.comp != RichComparer(6))

        self.assertTrue(self.comp != 7)
        self.assertTrue(self.comp != RichComparer(7))

    def testLessThan(self):
        self.assertTrue(self.comp < 7)
        self.assertTrue(self.comp < RichComparer(7))

        self.assertFalse(self.comp < 5)
        self.assertFalse(self.comp < RichComparer(5))

        self.assertFalse(self.comp < 6)
        self.assertFalse(self.comp < RichComparer(6))

    def testGreaterThan(self):
        self.assertTrue(self.comp > 5)
        self.assertTrue(self.comp > RichComparer(5))

        self.assertFalse(self.comp > 7)
        self.assertFalse(self.comp > RichComparer(7))

        self.assertFalse(self.comp > 6)
        self.assertFalse(self.comp > RichComparer(6))

    def testLessThanEqual(self):
        self.assertTrue(self.comp <= 7)
        self.assertTrue(self.comp <= RichComparer(7))
        self.assertTrue(self.comp <= 6)
        self.assertTrue(self.comp <= RichComparer(6))

        self.assertFalse(self.comp <= 5)
        self.assertFalse(self.comp <= RichComparer(5))

    def testGreaterThanEqual(self):
        self.assertTrue(self.comp >= 5)
        self.assertTrue(self.comp >= RichComparer(5))
        self.assertTrue(self.comp >= 6)
        self.assertTrue(self.comp >= RichComparer(6))

        self.assertFalse(self.comp >= 7)
        self.assertFalse(self.comp >= RichComparer(7))
        

class RichComparer(RichComparisonMixin):

    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        if not hasattr(other, 'value'):
            return self.value == other
        return self.value == other.value

    def __lt__(self, other):
        if not hasattr(other, 'value'):
            return self.value < other
        return self.value < other.value


class DummyIndex(object):

    value = None

    def index_doc(self, docid, value):
        value = self.discriminate(value, _marker)
        if value is _marker:
            # unindex the previous value
            self.unindex_doc(docid)
            # Store docid in set of unindexed docids
            self._not_indexed.add(docid)
            return None
        if docid in self._not_indexed:
            # Remove from set of unindexed docs if it was in there.
            self._not_indexed.remove(docid)
        self._docids.add(docid)
        self.value = value
        return value

    def unindex_doc(self, docid):
        _not_indexed = self._not_indexed
        if docid in _not_indexed:
            _not_indexed.remove(docid)
        if docid in self._docids:
            self._docids.remove(docid)
            self.unindexed = docid

    def indexed(self):
        return self._docids

    def not_indexed(self):
        return self._not_indexed

    def sort(self, ids, reverse, limit, sort_type=None, raise_unsortable=True):
        self.ids = ids
        self.reverse = reverse
        self.limit = limit
        self.sort_type = sort_type
        self.raise_unsortable = raise_unsortable
        return sorted(ids)
    



import sys

PY3 = sys.version_info[0] >= 3

if PY3: #pragma NO COVER
    def b(s):
        return s.encode("latin-1")
    def u(s):
        return s
    string_types = str,
    text_type = str
    binary_type = bytes
    integer_types = int,
    def non_native_string(x):
        if isinstance(x, bytes):
            return x
        return bytes(x, 'unicode_escape')
    def make_binary(x):
        if isinstance(x, bytes):
            return x
        return x.encode('ascii')
    intern = sys.intern
    xrange = range
    _long = int
    def _iteritems(x):
        return x.items()
    def _items(x):
        return list(x.items())
    def _maxint():
        return sys.maxsize
else: #pragma NO COVER
    def b(s):
        return s
    def u(s):
        return unicode(s, "unicode_escape")
    string_types = basestring,
    text_type = unicode
    binary_type = str
    integer_types = (int, long)
    def non_native_string(x):
        if isinstance(x, unicode):
            return x
        return unicode(x, 'unicode_escape')
    def make_binary(x):
        if isinstance(x, str):
            return x
        return x.encode('ascii')
    intern = intern
    xrange = xrange
    _long = long
    def _iteritems(x):
        return x.iteritems()
    def _items(x):
        return x.items()
    def _maxint():
        return sys.maxint



from zope.interface import implementer

from ..interfaces import (
    IIndex,
    IIndexStatistics,
    )
from ..util import BaseIndexMixin

from persistent import Persistent

from BTrees.Length import Length

from .interfaces import IKeywordQuerying
from .. import query
from .._compat import string_types

_marker = object()

@implementer(
    IIndex,
    IIndexStatistics,
    IKeywordQuerying,
    )
class KeywordIndex(BaseIndexMixin, Persistent):
    """
    Keyword index.

    Query types supported:

    - Eq

    - NotEq

    - In

    - NotIn

    - Any

    - NotAny

    - All

    - NotAll

    """

    # If a word is referenced by at least tree_threshold docids,
    # use a TreeSet for that word instead of a Set.
    tree_threshold = 64

    def __init__(self, discriminator, family=None):
        if family is not None:
            self.family = family
        if not callable(discriminator):
            if not isinstance(discriminator, string_types):
                raise ValueError('discriminator value must be callable or a '
                                 'string')
        self.discriminator = discriminator
        self.reset()

    def reset(self):
        """Initialize forward and reverse mappings."""

        # The forward index maps index keywords to a sequence of docids
        self._fwd_index = self.family.OO.BTree()

        # The reverse index maps a docid to its keywords
        # TODO: Using a vocabulary might be the better choice to store
        # keywords since it would allow use to use integers instead of strings
        self._rev_index = self.family.IO.BTree()
        self._num_docs = Length(0)
        self._not_indexed = self.family.IF.TreeSet()

    def unique_values(self):
        """ Return the unique values in the index for all docids as an iterable
        """
        return self._fwd_index.keys()
    
    def reindex_doc(self, docid, value):
        # the base index' index_doc method special-cases a reindex
        return self.index_doc(docid, value)

    def has_doc(self, docid):
        return bool(docid in self._rev_index)

    def indexed(self):
        return self._rev_index.keys()

    def not_indexed(self):
        return self._not_indexed

    def word_count(self):
        """Return the number of indexed words"""
        return len(self._fwd_index)

    def applyAny(self, values):
        return self.apply({'query': values, 'operator': 'or'})

    def any(self, value):
        return query.Any(self, value)

    def applyNotAny(self, *args, **kw):
        return self._negate(self.applyAny, *args, **kw)

    def notany(self, value):
        return query.NotAny(self, value)

    def applyAll(self, values):
        return self.apply({'query': values, 'operator': 'and'})

    def all(self, value):
        return query.All(self, value)

    def applyNotAll(self, *args, **kw):
        return self._negate(self.applyAll, *args, **kw)

    def notall(self, value):
        return query.NotAll(self, value)

    def applyEq(self, value):
        return self.apply([value])

    def eq(self, value):
        return query.Eq(self, value)

    def applyNotEq(self, *args, **kw):
        return self._negate(self.applyEq, *args, **kw)

    def noteq(self, value):
        return query.NotEq(self, value)

    def normalize(self, seq):
        """Perform normalization on sequence of keywords.

        Return normalized sequence. This method may be
        overriden by subclasses.

        """
        return seq

    def document_repr(self, docid, default=None):
        result = self._rev_index.get(docid, default)
        if result is not default:
            return repr(result)
        return default

    def index_doc(self, docid, obj):
        seq = self.discriminate(obj, _marker)

        if seq is _marker:
            if not (docid in self._not_indexed):
                # unindex the previous value
                self.unindex_doc(docid)
                # Store docid in set of unindexed docids
                self._not_indexed.add(docid)
            return None

        if docid in self._not_indexed:
            # Remove from set of unindexed docs if it was in there.
            self._not_indexed.remove(docid)

        if isinstance(seq, string_types):
            raise TypeError('seq argument must be a list/tuple of strings')

        old_kw = self._rev_index.get(docid, None)
        if not seq:
            if old_kw:
                self.unindex_doc(docid)
            return

        seq = self.normalize(seq)

        new_kw = self.family.OO.Set(seq)

        if old_kw is None:
            self._insert_forward(docid, new_kw)
            self._insert_reverse(docid, new_kw)
            self._num_docs.change(1)
        else:
            # determine added and removed keywords
            kw_added = self.family.OO.difference(new_kw, old_kw)
            kw_removed = self.family.OO.difference(old_kw, new_kw)

            if not (kw_added or kw_removed):
                return

            # removed keywords are removed from the forward index
            for word in kw_removed:
                fwd = self._fwd_index[word]
                fwd.remove(docid)
                if not fwd:
                    del self._fwd_index[word]

            # now update reverse and forward indexes
            self._insert_forward(docid, kw_added)
            self._insert_reverse(docid, new_kw)

    def unindex_doc(self, docid):
        _not_indexed = self._not_indexed
        if docid in _not_indexed:
            _not_indexed.remove(docid)
        
        idx  = self._fwd_index

        try:
            for word in self._rev_index[docid]:
                idx[word].remove(docid)
                if not idx[word]:
                    del idx[word]
        except KeyError:
            msg = 'WAAA!  Inconsistent'
            return

        try:
            del self._rev_index[docid]
        except KeyError: #pragma NO COVERAGE
            msg = 'WAAA!  Inconsistent'

        self._num_docs.change(-1)

    def _insert_forward(self, docid, words):
        """insert a sequence of words into the forward index """

        idx = self._fwd_index
        get_word_idx = idx.get
        IF = self.family.IF
        Set = IF.Set
        TreeSet = IF.TreeSet
        for word in words:
            word_idx = get_word_idx(word)
            if word_idx is None:
                idx[word] = word_idx = Set()
            word_idx.insert(docid)
            if (not isinstance(word_idx, TreeSet) and
                    len(word_idx) >= self.tree_threshold):
                # Convert to a TreeSet.
                idx[word] = TreeSet(word_idx)

    def _insert_reverse(self, docid, words):
        """ add words to forward index """

        if words:
            self._rev_index[docid] = words

    def search(self, query, operator='and'):
        """Execute a search given by 'query'."""
        if isinstance(query, string_types):
            query = [query]

        query = self.normalize(query)

        sets = []
        for word in query:
            docids = self._fwd_index.get(word, self.family.IF.Set())
            sets.append(docids)

        if operator == 'or':
            rs = self.family.IF.multiunion(sets)
        elif operator == 'and':
            # sort smallest to largest set so we intersect the smallest
            # number of document identifiers possible
            sets.sort(key=len)
            rs = None
            for set in sets:
                rs = self.family.IF.intersection(rs, set)
                if not rs:
                    break
        else:
            raise TypeError('Keyword index only supports `and` and `or` '
                            'operators, not `%s`.' % operator)

        if rs:
            return rs
        else:
            return self.family.IF.Set()

    def apply(self, query):
        operator = 'and'
        if isinstance(query, dict):
            if 'operator' in query:
                operator = query['operator']
            query = query['query']
        return self.search(query, operator=operator)

    def optimize(self):
        """Optimize the index. Call this after changing tree_threshold.

        This converts internal data structures between
        Sets and TreeSets based on tree_threshold.
        """
        idx = self._fwd_index
        IF = self.family.IF
        Set = IF.Set
        TreeSet = IF.TreeSet
        items = list(self._fwd_index.items())
        for word, word_idx in items:
            if len(word_idx) >= self.tree_threshold:
                if not isinstance(word_idx, TreeSet):
                    # Convert to a TreeSet.
                    idx[word] = TreeSet(word_idx)
            else:
                if isinstance(word_idx, TreeSet):
                    # Convert to a Set.
                    idx[word] = Set(word_idx)




##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Abstract base class for full text index with relevance ranking.
"""
import math

from hypatia._compat import u
from persistent import Persistent
from zope.interface import implementer

import BTrees
from BTrees import Length
from BTrees.IOBTree import IOBTree

from ..interfaces import (
    IIndexInjection,
    IIndexStatistics,
    )
from .interfaces import (
    IExtendedQuerying,
    ILexiconBasedIndex,
    )
from . import widcode
from .setops import (
    mass_weightedIntersection,
    mass_weightedUnion,
    )

@implementer(
    IIndexInjection,
    IIndexStatistics,
    ILexiconBasedIndex,
    IExtendedQuerying
    )
class BaseIndex(Persistent):

    family = BTrees.family64

    lexicon = property(lambda self: self._lexicon,)

    def __init__(self, lexicon, family=None):
        if family is not None:
            self.family = family
        self._lexicon = lexicon
        self.reset()

    def reset(self):
        # wid -> {docid -> weight}; t -> D -> w(D, t)
        # Different indexers have different notions of term weight, but we
        # expect each indexer to use ._wordinfo to map wids to its notion
        # of a docid-to-weight map.
        # There are two kinds of OOV words:  wid 0 is explicitly OOV,
        # and it's possible that the lexicon will return a non-zero wid
        # for a word we don't currently know about.  For example, if we
        # unindex the last doc containing a particular word, that wid
        # remains in the lexicon, but is no longer in our _wordinfo map;
        # lexicons can also be shared across indices, and some other index
        # may introduce a lexicon word we've never seen.
        # A word is in-vocabulary for this index if and only if
        # wid in _wordinfo.  Note that wid 0 must not be a key.
        # This does not use the BTree family since wids are always "I"
        # flavor trees.
        self._wordinfo = IOBTree()

        # docid -> weight
        # Different indexers have different notions of doc weight, but we
        # expect each indexer to use ._docweight to map docids to its
        # notion of what a doc weight is.
        self._docweight = self.family.IF.BTree()

        # docid -> WidCode'd list of wids
        # Used for un-indexing, and for phrase search.
        self._docwords = self.family.IO.BTree()

        # Use a BTree length for efficient length computation w/o conflicts
        self.word_count = Length.Length()
        self.indexed_count = Length.Length()

    def word_count(self):
        """Return the number of words in the index."""
        # This must be overridden by subclasses which do not set the
        # attribute on their instances.
        raise NotImplementedError

    def indexed_count(self):
        """Return the number of documents in the index."""
        # This must be overridden by subclasses which do not set the
        # attribute on their instances.
        raise NotImplementedError

    def get_words(self, docid):
        """Return a list of the wordids for a given docid."""
        return widcode.decode(self._docwords[docid])

    def document_repr(self, docid, default=None):
        try:
            words = []
            wids = self.get_words(docid)
            for wid in wids:
                words.append(self._lexicon.get_word(wid))
            return u(' ').join(words)
        except KeyError:
            return default

    # A subclass may wish to extend or override this.
    def index_doc(self, docid, text):
        if docid in self._docwords:
            return self.reindex_doc(docid, text)
        wids = self._lexicon.sourceToWordIds(text)
        wid2weight, docweight = self._get_frequencies(wids)
        self._mass_add_wordinfo(wid2weight, docid)
        self._docweight[docid] = docweight
        self._docwords[docid] = widcode.encode(wids)
        try:
            self.indexed_count.change(1)
        except AttributeError:
            # upgrade indexed_count to Length object
            self.indexed_count = Length.Length(len(self._docweight))
        return len(wids)

    # A subclass may wish to extend or override this.  This is for adjusting
    # to a new version of a doc that already exists.  The goal is to be
    # faster than simply unindexing the old version in its entirety and then
    # adding the new version in its entirety.
    def reindex_doc(self, docid, text):
        # Touch as few docid->w(docid, score) maps in ._wordinfo as possible.
        old_wids = self.get_words(docid)
        old_wid2w, old_docw = self._get_frequencies(old_wids)

        new_wids = self._lexicon.sourceToWordIds(text)
        new_wid2w, new_docw = self._get_frequencies(new_wids)

        old_widset = self.family.IF.TreeSet(old_wid2w.keys())
        new_widset = self.family.IF.TreeSet(new_wid2w.keys())

        IF = self.family.IF
        in_both_widset = IF.intersection(old_widset, new_widset)
        only_old_widset = IF.difference(old_widset, in_both_widset)
        only_new_widset = IF.difference(new_widset, in_both_widset)
        del old_widset, new_widset

        for wid in only_old_widset.keys():
            self._del_wordinfo(wid, docid)

        for wid in only_new_widset.keys():
            self._add_wordinfo(wid, new_wid2w[wid], docid)

        for wid in in_both_widset.keys():
            # For the Okapi indexer, the "if" will trigger only for words
            # whose counts have changed.  For the cosine indexer, the "if"
            # may trigger for every wid, since W(d) probably changed and
            # W(d) is divided into every score.
            newscore = new_wid2w[wid]
            if old_wid2w[wid] != newscore:
                self._add_wordinfo(wid, newscore, docid)

        self._docweight[docid] = new_docw
        self._docwords[docid] = widcode.encode(new_wids)
        return len(new_wids)

    # Subclass must override.
    def _get_frequencies(self, wids):
        # Compute term frequencies and a doc weight, whatever those mean
        # to an indexer.
        # Return pair:
        #    {wid0: w(d, wid0), wid1: w(d, wid1),  ...],
        #    docweight
        # The wid->weight mappings are fed into _add_wordinfo, and docweight
        # becomes the value of _docweight[docid].
        raise NotImplementedError

    def has_doc(self, docid):
        return docid in self._docwords

    # A subclass may wish to extend or override this.
    def unindex_doc(self, docid):
        if docid not in self._docwords:
            return
        for wid in self.family.IF.TreeSet(self.get_words(docid)).keys():
            self._del_wordinfo(wid, docid)
        del self._docwords[docid]
        del self._docweight[docid]
        try:
            self.indexed_count.change(-1)
        except AttributeError:
            # upgrade indexed_count to Length object
            self.indexed_count = Length.Length(len(self._docweight))

    def search(self, term):
        wids = self._lexicon.termToWordIds(term)
        if not wids:
            return None # All docs match
        wids = self._remove_oov_wids(wids)
        return mass_weightedUnion(self._search_wids(wids), self.family)

    def search_glob(self, pattern):
        wids = self._lexicon.globToWordIds(pattern)
        wids = self._remove_oov_wids(wids)
        return mass_weightedUnion(self._search_wids(wids), self.family)

    def search_phrase(self, phrase):
        wids = self._lexicon.termToWordIds(phrase)
        cleaned_wids = self._remove_oov_wids(wids)
        if len(wids) != len(cleaned_wids):
            # At least one wid was OOV:  can't possibly find it.
            return self.family.IF.BTree()
        scores = self._search_wids(wids)
        hits = mass_weightedIntersection(scores, self.family)
        if not hits:
            return hits
        code = widcode.encode(wids)
        result = self.family.IF.BTree()
        for docid, weight in hits.items():
            docwords = self._docwords[docid]
            if docwords.find(code) >= 0:
                result[docid] = weight
        return result

    def _remove_oov_wids(self, wids):
        return [wid for wid in wids if wid in self._wordinfo]

    # Subclass must override.
    # The workhorse.  Return a list of (IFBucket, weight) pairs, one pair
    # for each wid t in wids.  The IFBucket, times the weight, maps D to
    # TF(D,t) * IDF(t) for every docid D containing t.  wids must not
    # contain any OOV words.
    def _search_wids(self, wids):
        raise NotImplementedError

    # Subclass must override.
    # It's not clear what it should do.  It must return an upper bound on
    # document scores for the query.  It would be nice if a document score
    # divided by the query's query_weight gave the proabability that a
    # document was relevant, but nobody knows how to do that.  For
    # CosineIndex, the ratio is the cosine of the angle between the document
    # and query vectors.  For OkapiIndex, the ratio is a (probably
    # unachievable) upper bound with no "intuitive meaning" beyond that.
    def query_weight(self, terms):
        raise NotImplementedError

    DICT_CUTOFF = 10

    def _add_wordinfo(self, wid, f, docid):
        # Store a wordinfo in a dict as long as there are less than
        # DICT_CUTOFF docids in the dict.  Otherwise use an IFBTree.

        # The pickle of a dict is smaller than the pickle of an
        # IFBTree, substantially so for small mappings.  Thus, we use
        # a dictionary until the mapping reaches DICT_CUTOFF elements.

        # The cutoff is chosen based on the implementation
        # characteristics of Python dictionaries.  The dict hashtable
        # always has 2**N slots and is resized whenever it is 2/3s
        # full.  A pickled dict with 10 elts is half the size of an
        # IFBTree with 10 elts, and 10 happens to be 2/3s of 2**4.  So
        # choose 10 as the cutoff for now.

        # The IFBTree has a smaller in-memory representation than a
        # dictionary, so pickle size isn't the only consideration when
        # choosing the threshold.  The pickle of a 500-elt dict is 92%
        # of the size of the same IFBTree, but the dict uses more
        # space when it is live in memory.  An IFBTree stores two C
        # arrays of ints, one for the keys and one for the values.  It
        # holds up to 120 key-value pairs in a single bucket.
        doc2score = self._wordinfo.get(wid)
        if doc2score is None:
            doc2score = {} # XXX Holy ConflictError, Batman!
            try:
                self.word_count.change(1)
            except AttributeError:
                # upgrade word_count to Length object
                self.word_count = Length.Length(len(self._wordinfo))
                self.word_count.change(1)
        else:
            # _add_wordinfo() is called for each update.  If the map
            # size exceeds the DICT_CUTOFF, convert to an IFBTree.
            # Obscure:  First check the type.  If it's not a dict, it
            # can't need conversion, and then we can avoid an expensive
            # len(IFBTree).
            if (isinstance(doc2score, type({})) and
                len(doc2score) == self.DICT_CUTOFF):
                doc2score = self.family.IF.BTree(doc2score)
        doc2score[docid] = f
        self._wordinfo[wid] = doc2score # not redundant:  Persistency!

    #    self._mass_add_wordinfo(wid2weight, docid)
    #
    # is the same as
    #
    #    for wid, weight in wid2weight.items():
    #        self._add_wordinfo(wid, weight, docid)
    #
    # except that _mass_add_wordinfo doesn't require so many function calls.
    def _mass_add_wordinfo(self, wid2weight, docid):
        dicttype = type({})
        get_doc2score = self._wordinfo.get
        new_word_count = 0
        for wid, weight in wid2weight.items():
            doc2score = get_doc2score(wid)
            if doc2score is None:
                doc2score = {}
                new_word_count += 1
            elif (isinstance(doc2score, dicttype) and
                  len(doc2score) == self.DICT_CUTOFF):
                doc2score = self.family.IF.BTree(doc2score)
            doc2score[docid] = weight
            self._wordinfo[wid] = doc2score # not redundant:  Persistency!
        try:
            self.word_count.change(new_word_count)
        except AttributeError:
            # upgrade word_count to Length object
            self.word_count = Length.Length(len(self._wordinfo))

    def _del_wordinfo(self, wid, docid):
        doc2score = self._wordinfo[wid]
        del doc2score[docid]
        if doc2score:
            self._wordinfo[wid] = doc2score # not redundant:  Persistency!
        else:
            del self._wordinfo[wid]
            try:
                self.word_count.change(-1)
            except AttributeError:
                # upgrade word_count to Length object
                self.word_count = Length.Length(len(self._wordinfo))

def inverse_doc_frequency(term_count, num_items):
    """Return the inverse doc frequency for a term,

    that appears in term_count items in a collection with num_items
    total items.
    """
    # implements IDF(q, t) = log(1 + N/f(t))
    return math.log(1.0 + float(num_items) / term_count)



import itertools
import BTrees

from persistent import Persistent
from ZODB.broken import Broken
from zope.interface import implementer

_marker = object()

from .. import exc
from ..interfaces import (
    IResultSet,
    STABLE,
    )

@implementer(IResultSet)
class ResultSet(object):
    """Implements :class:`hypatia.interfaces.IResultSet`"""

    family = BTrees.family64

    def __init__(self, ids, numids, resolver, sort_type=None):
        self.ids = ids # only guaranteed to be iterable, not sliceable
        self.numids = numids
        self.resolver = resolver
        self.sort_type = sort_type
        
    def __len__(self):
        return self.numids

    def sort(self, index, reverse=False, limit=None, sort_type=None,
             raise_unsortable=True):
        if sort_type is None:
            sort_type = self.sort_type
        
        ids = self.ids

        if not hasattr(ids, '__len__'):
            # indexes have no obligation to be able to sort generators
            ids = list(ids)
            self.ids = ids

        ids = index.sort(
            self.ids,
            reverse=reverse,
            limit=limit,
            sort_type=sort_type,
            raise_unsortable=raise_unsortable,
            )

        numids = self.numids

        if limit:
            numids = min(numids, limit)

        return self.__class__(ids, numids, self.resolver, sort_type=STABLE)

    def first(self, resolve=True):
        # return the first object or None
        resolver = self.resolver
        if resolver is None or not resolve:
            for id_ in self.ids:
                # if self.ids is not a list or a tuple, allow this result set
                # to be iterated after first() is called and allow first() to
                # be idempotent
                if not hasattr(self.ids, '__len__'):
                    self.ids = itertools.chain([id_], self.ids)
                return id_
        else:
            for id_ in self.ids:
                # if self.ids is not a list or a tuple, allow this result set
                # to be iterated after first() is called and allow first() to
                # be idempotent
                if not hasattr(self.ids, '__len__'):
                    self.ids = itertools.chain([id_], self.ids)
                return resolver(id_)

    def one(self, resolve=True):
        if self.numids == 1:
            return self.first(resolve=resolve)
        if self.numids > 1:
            raise exc.MultipleResults(self)
        else:
            raise exc.NoResults(self)

    def _resolve_all(self, resolver):
        for id_ in self.ids:
            yield resolver(id_)

    def all(self, resolve=True):
        resolver = self.resolver
        if resolver is None or not resolve:
            return self.ids
        else:
            return self._resolve_all(resolver)

    def __iter__(self):
        return iter(self.all())

    def intersect(self, docids):
        """ Intersect this resultset with a sequence of docids or
        another resultset.  Returns a new ResultSet. """
        # NB: we can't use an intersection function here because
        # self.ids may be a generator
        if isinstance(docids, ResultSet):
            docids = docids.ids
        filtered_ids = [ x for x in self.ids if x in docids ]
        return self.__class__(filtered_ids, len(filtered_ids), self.resolver)

class BaseIndexMixin(object):
    """ Mixin class for indexes that implements common behavior """

    family = BTrees.family64

    def discriminate(self, obj, default):
        """ See interface IIndexInjection """
        if callable(self.discriminator):
            value = self.discriminator(obj, _marker)
        else:
            value = getattr(obj, self.discriminator, _marker)

        if value is _marker:
            return default
        
        if isinstance(value, Persistent):
            raise ValueError('Catalog cannot index persistent object %s' %
                             value)

        if isinstance(value, Broken):
            raise ValueError('Catalog cannot index broken object %s' %
                             value)

        return value

    def reindex_doc(self, docid, obj):
        """ See interface IIndexInjection """
        self.unindex_doc(docid)
        self.index_doc(docid, obj)

    def indexed_count(self):
        """ See IIndexedDocuments """
        return len(self.indexed())

    def not_indexed_count(self):
        """ See IIndexedDocuments """
        return len(self.not_indexed())

    def docids(self):
        """ See IIndexedDocuments """
        not_indexed = self.not_indexed()
        indexed = self.indexed()
        if len(not_indexed) == 0:
            return self.family.IF.Set(indexed)
        elif len(indexed) == 0:
            return not_indexed
        indexed = self.family.IF.Set(indexed)
        return self.family.IF.union(not_indexed, indexed)

    def docids_count(self):
        """ See IIndexedDocuments """
        return len(self.docids())

    def apply_intersect(self, query, docids):
        """ Default apply_intersect implementation """
        result = self.apply(query)
        if docids is None:
            return result
        return self.family.IF.weightedIntersection(result, docids)[1]

    def _negate(self, apply_func, *args, **kw):
        positive = apply_func(*args, **kw)
        all = self.docids()
        if len(positive) == 0:
            return all
        return self.family.IF.difference(all, positive)

    def qname(self):
        # used in query representations; __name__ should be set by
        # catalog __setitem__ but if it's not, we fall back to a generic
        # representation
        return getattr(
            self,
            '__name__',
            str(self),
            )
        
    def resultset_from_query(self, query, names=None, resolver=None):
        # default resultset factory; meant to be overridden by systems that
        # have a default resolver.  NB: although the default implementation
        # below does not access "self", so it would appear that this could be
        # turned into a classmeth or staticmethod, subclasses that override may
        # expect self, so this is a plain method.
        docids = query._apply(names)
        numdocs = len(docids)
        return ResultSet(docids, numdocs, resolver)

    def flush(self, *arg, **kw):
        """ Hookable by upstream systems"""
        pass

class RichComparisonMixin(object):
    # Stolen from http://www.voidspace.org.uk/python/recipebook.shtml#comparison

    def __eq__(self, other):
        raise NotImplementedError("Equality not implemented")

    def __lt__(self, other):
        raise NotImplementedError("Less than not implemented")

    def __ne__(self, other):
        return not self.__eq__(other)
    
    def __gt__(self, other):
        return not (self.__lt__(other) or self.__eq__(other))

    def __le__(self, other):
        return self.__eq__(other) or self.__lt__(other)

    def __ge__(self, other):
        return self.__eq__(other) or self.__gt__(other)




##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Text Index Tests
"""
import unittest

_marker = object()

class Test_widcode(unittest.TestCase):

    def test_encode_1_to_7_bits(self):
        from ..._compat import xrange
        from ..widcode import encode
        for wid in xrange(2**7):
            code = encode([wid])
            self.assertEqual(code, chr(wid + 128))

    def test_encode_8_to_14_bits(self):
        from ..._compat import xrange
        from ..widcode import encode
        for wid in xrange(2**7, 2**14):
            hi, lo = divmod(wid, 128)
            code = encode([wid])
            self.assertEqual(code, chr(hi + 128) + chr(lo))

    def test_encode_15_to_21_bits(self):
        from ..._compat import xrange
        from ..widcode import encode
        for wid in xrange(2**14, 2**21, 255):
            mid, lo = divmod(wid, 128)
            hi, mid = divmod(mid, 128)
            code = encode([wid])
            self.assertEqual(code, chr(hi + 128) + chr(mid) + chr(lo))

    def test_encode_22_to_28_bits(self):
        from ..._compat import xrange
        from ..widcode import encode
        STEP = (256 * 512) - 1
        for wid in xrange(2**21, 2**28, STEP):
            lmid, lo = divmod(wid, 128)
            hmid, lmid = divmod(lmid, 128)
            hi, hmid = divmod(hmid, 128)
            code = encode([wid])
            self.assertEqual(code,
                             chr(hi + 128) + chr(hmid) + chr(lmid) + chr(lo))

    def test_decode_zero(self):
        from ..widcode import decode
        self.assertEqual(decode('\x80'), [0])

    def test__decode_other_one_byte_asserts(self):
        from ..widcode import _decode
        for wid in range(1, 128):
            self.assertRaises(AssertionError, _decode, chr(128+wid))

    def test__decode_two_bytes_asserts(self):
        from ..widcode import _decode
        for wid in range(128, 2**14):
            hi, lo = divmod(wid, 128)
            code = chr(hi + 128) + chr(lo)
            self.assertRaises(AssertionError, _decode, code)
                
    def test__decode_three_bytes(self):
        from ..widcode import _decode
        for wid in range(2**14, 2**21, 247):
            mid, lo = divmod(wid, 128)
            hi, mid = divmod(mid, 128)
            code = chr(hi + 128) + chr(mid) + chr(lo)
            self.assertEqual(_decode(code), wid)

    def test__decode_four_bytes(self):
        from ..widcode import _decode
        STEP = (256 * 512) - 7
        for wid in range(2**21, 2**28, STEP):
            lmid, lo = divmod(wid, 128)
            hmid, lmid = divmod(lmid, 128)
            hi, hmid = divmod(hmid, 128)
            code = chr(hi + 128) + chr(hmid) + chr(lmid) + chr(lo)
            self.assertEqual(_decode(code), wid)

    def test_symmetric(self):
        from ..._compat import xrange
        from ..widcode import decode
        from ..widcode import encode
        for wid in xrange(2**14, 2**21, 247):
            wids = [wid]
            code = encode(wids)
            self.assertEqual(decode(code), wids)




import operator

import BTrees
from persistent.mapping import PersistentMapping
from zope.interface import implementer

from ..interfaces import ICatalog
from ..interfaces import ICatalogQuery
from ..query import parse_query
from .._compat import string_types

@implementer(ICatalog)
class Catalog(PersistentMapping):

    family = BTrees.family64

    def __init__(self, family=None):
        PersistentMapping.__init__(self)
        if family is not None:
            self.family = family

    def __setitem__(self, name, index):
        index.__name__ = name
        return PersistentMapping.__setitem__(self, name, index)

    def reset(self):
        """ Clear all indexes in this catalog. """
        for index in self.values():
            index.reset()

    def index_doc(self, docid, obj):
        """Register the document represented by ``obj`` in indexes of
        this catalog using docid ``docid``."""
        assertint(docid)
        for index in self.values():
            index.index_doc(docid, obj)

    def unindex_doc(self, docid):
        """Unregister the document id from indexes of this catalog."""
        assertint(docid)
        for index in self.values():
            index.unindex_doc(docid)

    def reindex_doc(self, docid, obj):
        """ Reindex the document referenced by docid using the object
        passed in as ``obj`` (typically just does the equivalent of
        ``unindex_doc``, then ``index_doc``, but specialized indexes
        can override the method that this API calls to do less work. """
        assertint(docid)
        for index in self.values():
            index.reindex_doc(docid, obj)

def assertint(docid):
    if not isinstance(docid, int):
        raise ValueError('%r is not an integer value; document ids must be '
                         'integers' % docid)


@implementer(ICatalogQuery)
class CatalogQuery(object):
    """ Legacy query API for non-index-based queries; might be useful if/when
    an index-based query doesn't work properly, or a particular contstraint
    can't be spelled with one."""

    family = BTrees.family64
    
    def __init__(self, catalog, family=None):
        self.catalog = catalog
        if family is not None:
            self.family = family

    def sort(self, docidset, sort_index, limit=None, sort_type=None,
             reverse=False):
        """ Return ``(num, sorted-resultseq)`` for the concrete docidset. """

        result = docidset
        numdocs = len(docidset)

        if sort_index:
            index = self.catalog[sort_index]
            result = index.sort(
                result, reverse=reverse, limit=limit, sort_type=sort_type
                )
            if limit:
                numdocs = min(numdocs, limit)
            return numdocs, result
        else:
            return numdocs, result

    def search(self, **query):
        """ Use the query terms to perform a query.  Return a tuple of
        (num, resultseq) based on the merging of results from
        individual indexes.

        .. note::

           This method is deprecated. Use
           :func:`hypatia.catalog.CatalogQuery.__call__` instead.


        """
        sort_index = None
        reverse = False
        limit = None
        sort_type = None
        index_query_order = None

        if 'sort_index' in query:
            sort_index = query.pop('sort_index')
        if 'reverse' in query:
            reverse = query.pop('reverse')
        if 'limit' in query:
            limit = query.pop('limit')
        if 'sort_type' in query:
            sort_type = query.pop('sort_type')
        if 'index_query_order' in query:
            index_query_order = query.pop('index_query_order')

        if index_query_order is None:
            # unordered query (use apply)
            results = []
            for index_name, index_query in query.items():
                index = self.catalog.get(index_name)
                if index is None:
                    raise ValueError('No such index %s' % index_name)
                r = index.apply(index_query)
                if not r:
                    # empty results, bail early; intersect will be null
                    return 0, r

                results.append((len(r), r))

            if not results:
                return 0, ()

            # order smallest to largest
            _, result = sorted(results,
                               key=operator.itemgetter(0),
                              ).pop(0)
            for _, r in results:
                _, result = self.family.IF.weightedIntersection(result, r)

            if not result:
                return 0, ()

        else:
            # ordered query (use apply_intersect)
            result = None
            _marker = object()
            for index_name in index_query_order:
                index_query = query.get(index_name, _marker)
                if index_query is _marker:
                    continue
                index = self.catalog.get(index_name)
                if index is None:
                    raise ValueError('No such index %s' % index_name)
                result = index.apply_intersect(index_query, result)
                if not result:
                    # empty results
                    return 0, result

        return self.sort(result, sort_index, limit, sort_type, reverse)

    def query(self, queryobject, sort_index=None, limit=None, sort_type=None,
              reverse=False, names=None):
        """ Use the arguments to perform a query.  Return a tuple of
        (num, resultseq)."""
        if isinstance(queryobject, string_types):
            queryobject = parse_query(queryobject, self.catalog)
        results = queryobject._apply(names)
        return self.sort(results, sort_index, limit, sort_type, reverse)

    __call__ = query




import unittest

class TestBadResults(unittest.TestCase):
    def _makeOne(self, resultset):
        from hypatia.exc import BadResults
        return BadResults([1,2,3])

    def test_it(self):
        inst = self._makeOne([1,2,3])
        self.assertEqual(inst.resultset, [1,2,3])

class TestUnsortable(unittest.TestCase):
    def _makeOne(self, docids):
        from hypatia.exc import Unsortable
        return Unsortable(docids)

    def test___repr__(self):
        inst = self._makeOne([1,2,3])
        self.assertEqual(repr(inst), '[1, 2, 3]')

    def test___str__(self):
        inst = self._makeOne([1,2,3])
        self.assertEqual(str(inst), '[1, 2, 3]')
        



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import os
import sys

from setuptools import setup, find_packages, Extension
from distutils.command.build_ext import build_ext
from distutils.errors import CCompilerError
from distutils.errors import DistutilsExecError
from distutils.errors import DistutilsPlatformError

class optional_build_ext(build_ext):
    """This class subclasses build_ext and allows
       the building of C extensions to fail.
    """
    def run(self):
        try:
            build_ext.run(self)
        
        except DistutilsPlatformError as e:
            self._unavailable(e)

    def build_extension(self, ext):
       try:
           build_ext.build_extension(self, ext)
        
       except (CCompilerError, DistutilsExecError) as e:
           self._unavailable(e)

    def _unavailable(self, e):
        def errprint(x):
            sys.stderr.write('%s\n' % x)
        errprint('*' * 80)
        errprint("""WARNING:

        An optional code optimization (C extension) could not be compiled.

        Optimizations for this package will not be available!""")
        errprint('')
        errprint(e)
        errprint('*' * 80)

try:
    here = os.path.abspath(os.path.dirname(__file__))

    with open(os.path.join(here, 'README.rst')) as f:
        README = f.read()

    with open(os.path.join(here, 'CHANGES.rst')) as f:
        CHANGES = f.read()
except IOError:
    README = CHANGES = ''

install_requires = [
    'persistent',
    'BTrees',
    'ZODB>=4.0.0b2', # py3 compat
    'zope.interface',
    ]

testing_extras = ['pytest', 'coverage']

docs_extras = [
    'Sphinx >= 3.0.0',  # Force RTD to use >= 3.0.0
]


setup(name='hypatia',
      version='0.4',
      description='Python package for searching and indexing',
      long_description=README + '\n\n' +  CHANGES,
      classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        ],
      keywords='indexing catalog search',
      author="Zope Foundation and Contributors",
      author_email="pylons-discuss@googlegroups.com",
      url="http://pylonsproject.org",
      license="ZPL 2.1",
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      tests_require = install_requires,
      install_requires = install_requires,
      extras_require = {
        'benchmark': ['PyChart'],
        'testing': testing_extras,
        'docs': docs_extras,
        },
      ext_modules=[
          Extension('hypatia.text.okascore',
              [os.path.join('hypatia', 'text', 'okascore.c')]),
      ],
      cmdclass = {'build_ext':optional_build_ext},
      test_suite="hypatia",
      ## entry_points = """\
      ## [console_scripts]
      ## catalog_benchmark = hypatia.benchmark.benchmark:run
      ## sortbench = hypatia.benchmark.sortbench:main
      ## """,
      )




class BadResults(Exception):
    """ Superclass of :exc:`hypatia.exc.MultipleResults` and
    :exc:`hypatia.exc.NoResults`.  Has an attribute named ``resultset`` which
    is the resultset related to the error."""
    def __init__(self, resultset):
        self.resultset = resultset

class MultipleResults(BadResults):
    """ Raised when a method that was expected to return a single result
    returns multiple results."""

class NoResults(BadResults):
    """ Raised when a method that was expected to return at least one result
    returns zero results."""

class Unsortable(Exception):
    """ Raised when a method which was expected to sort a set of provided
    document identifiers cannot sort one or more of those identifiers.  An
    attribute named ``docids`` is a sequence containing the identifiers. """
    def __init__(self, docids):
        self.docids = docids

    def __repr__(self):
        return '%r' % (list(self.docids),)
    __str__ = __repr__
        



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Text index.
"""
from persistent import Persistent
from zope.interface import implementer

from hypatia.interfaces import (
    IIndex,
    IIndexStatistics,
    IIndexSort,
    )

from .lexicon import (
    CaseNormalizer,
    Lexicon,
    Splitter,
    StopWordRemover,
    )
from .okapiindex import OkapiIndex
from .queryparser import QueryParser
from .parsetree import ParseError

from ..util import BaseIndexMixin 
from .. import query
from .._compat import string_types
from .._compat import _iteritems
from .._compat import _maxint

_marker = object()

@implementer(
    IIndex,
    IIndexSort,
    IIndexStatistics
    )
class TextIndex(BaseIndexMixin, Persistent):
    def __init__(self, discriminator, lexicon=None, index=None,
                 family=None):
        if family is not None:
            self.family = family
        if not callable(discriminator):
            if not isinstance(discriminator, string_types):
                raise ValueError('discriminator value must be callable or a '
                                 'string')
        self.discriminator = discriminator
        _explicit_lexicon = True
        if lexicon is None:
            _explicit_lexicon = False
            lexicon = Lexicon(Splitter(), CaseNormalizer(), StopWordRemover())
        if index is None:
            index = OkapiIndex(lexicon, family=self.family) # override family
        self.lexicon = _explicit_lexicon and lexicon or index.lexicon
        self.index = index
        self.reset()

    def reset(self):
        self._not_indexed = self.family.IF.TreeSet()
        self.index.reset()

    def document_repr(self, docid, default=None):
        return self.index.document_repr(docid, default)

    def index_doc(self, docid, obj):
        text = self.discriminate(obj, _marker)

        if text is _marker:
            # unindex the previous value
            self.unindex_doc(docid)
            # Store docid in set of unindexed docids
            self._not_indexed.add(docid)
            return None

        if docid in self._not_indexed:
            # Remove from set of unindexed docs if it was in there.
            self._not_indexed.remove(docid)

        self.index.index_doc(docid, text)

    def unindex_doc(self, docid):
        _not_indexed = self._not_indexed
        if docid in _not_indexed:
            _not_indexed.remove(docid)
        self.index.unindex_doc(docid)

    def reindex_doc(self, docid, object):
        # index_doc knows enough about reindexing to do the right thing
        return self.index_doc(docid, object)

    def indexed(self):
        return self.index._docwords.keys()

    def indexed_count(self):
        return self.index.indexed_count()

    def not_indexed(self):
        return self._not_indexed

    def word_count(self):
        """Return the number of words in the index."""
        return self.index.word_count()

    def parse_query(self, querytext):
        parser = QueryParser(self.lexicon)
        tree = parser.parseQuery(querytext)
        return tree

    def check_query(self, querytext):
        """ Returns True if the querytext can be turned into a parse tree,
        returns False if otherwise. """
        try:
            self.parse_query(querytext)
            return True
        except ParseError:
            return False

    def apply(self, querytext, start=0, count=None):
        tree = self.parse_query(querytext)
        results = tree.executeQuery(self.index)
        if results:
            qw = self.index.query_weight(tree.terms())
            
            # Hack to avoid ZeroDivisionError
            if qw == 0:
                qw = 1.0

            qw *= 1.0

            for docid, score in _iteritems(results):
                try:
                    results[docid] = score/qw
                except TypeError:
                    # We overflowed the score, perhaps wildly unlikely.
                    # Who knows.
                    results[docid] = _maxint() / 10.0

        return results
 
    def applyContains(self, value):
        return self.apply(value)

    def contains(self, value):
        return query.Contains(self, value)

    def applyNotContains(self, *args, **kw):
        return self._negate(self.applyContains, *args, **kw)

    def notcontains(self, value):
        return query.NotContains(self, value)

    applyEq = applyContains
    eq = contains

    applyNotEq = applyNotContains
    noteq = notcontains

    def sort(self, result, reverse=False, limit=None, sort_type=None,
             raise_unsortable=True):
        """Sort by text relevance.

        This only works if the query includes at least one text query,
        leading to a weighted result.  This method raises TypeError
        if the result is not weighted.

        A weighted result is a dictionary-ish object that has docids
        as keys and floating point weights as values.  This method
        sorts the dictionary by weight and returns the sorted
        docids as a list.
        """
        if not result:
            return result

        if not hasattr(result, 'items'):
            raise TypeError(
                "Unable to sort by relevance because the search "
                "result does not contain weights. To produce a weighted "
                "result, include a text search in the query.")

        items = [(weight, docid) for (docid, weight) in result.items()]
        # when reverse is false, output largest weight first.
        # when reverse is true, output smallest weight first.
        items.sort(reverse=not reverse)
        result = [docid for (weight, docid) in items]
        if limit:
            result = result[:limit]
        return result
    
